/**
 * Root class for all errors thrown by this library.
 * This should not be thrown directly, but rather one of its subclasses.
 */
export declare class TdfError extends Error {
    name: string;
    constructor(message?: string, cause?: Error);
}
/**
 * Errors that indicate the client or method does not have valid options.
 */
export declare class ConfigurationError extends TdfError {
    name: string;
}
/**
 * The assigned data attribute is not in the correct form.
 */
export declare class AttributeValidationError extends ConfigurationError {
    name: string;
    attribute: unknown;
    constructor(message: string, attribute: unknown, cause?: Error);
}
/**
 * Errors that indicate the TDF object is corrupt, invalid, or fails validation or decrypt.
 */
export declare class InvalidFileError extends TdfError {
}
/**
 * Indicates a decrypt failure, either due to an incorrect key, corrupt ciphertext, or inappropriate key parameters.
 */
export declare class DecryptError extends InvalidFileError {
    name: string;
}
export declare class IntegrityError extends InvalidFileError {
    name: string;
}
/**
 * Thrown when a KAS URL found in one or more required key access objects are not in the list of known and allowed KASes in the client.
 * This may indicate a malicious file - e.g. an attempt to DDoS a server by listing it as the KAS for many files, or to siphon credentials using a lookalike URL.
 */
export declare class UnsafeUrlError extends InvalidFileError {
    name: string;
    readonly url: string[];
    constructor(message: string, ...url: string[]);
}
/**
 * A network error (no response) from rewrap or other endpoint, Possibly fixed by retrying or adjusting your network settings; could indicate network failure.
 */
export declare class NetworkError extends TdfError {
    name: string;
}
/**
 * The service reports an unexpected error on its behalf, or a subcomponent (5xx).
 */
export declare class ServiceError extends TdfError {
    name: string;
}
/** Authentication failure (401) */
export declare class UnauthenticatedError extends TdfError {
    name: string;
}
/** Authorization failure (403) */
export declare class PermissionDeniedError extends TdfError {
    name: string;
}
/**
 * Version of file is unsupported, or file uses a feature that is not supported by this version of the library.
 */
export declare class UnsupportedFeatureError extends TdfError {
    name: string;
}
//# sourceMappingURL=errors.d.ts.map