import { Client, NanoTDF } from './nanotdf/index.js';
import { ClientConfig } from './nanotdf/Client.js';
export type EncryptOptions = {
    ecdsaBinding: boolean;
};
/**
 * NanoTDF SDK Client. Deprecated in favor of OpenTDF.
 *
 */
export declare class NanoTDFClient extends Client {
    /**
     * Decrypt ciphertext
     *
     * Pass a base64 string, TypedArray, or ArrayBuffer ciphertext and get a promise which resolves plaintext
     *
     * @param ciphertext Ciphertext to decrypt
     */
    decrypt(ciphertext: string | ArrayBufferLike): Promise<ArrayBuffer>;
    /**
     * Decrypt ciphertext of the legacy TDF, with the older, smaller i.v. calculation.
     *
     * Pass a base64 string, TypedArray, or ArrayBuffer ciphertext and get a promise which resolves plaintext
     *
     * @param ciphertext Ciphertext to decrypt
     */
    decryptLegacyTDF(ciphertext: string | ArrayBufferLike): Promise<ArrayBuffer>;
    /**
     * Encrypts the given data using the NanoTDF encryption scheme.
     *
     * @param data The data to be encrypted.
     * @param options The encryption options (currently unused).
     * @returns A promise that resolves to the encrypted data as an ArrayBuffer.
     * @throws If the initialization vector is not a number.
     */
    encrypt(data: string | ArrayBufferLike, options?: EncryptOptions): Promise<ArrayBuffer>;
}
export type DatasetConfig = ClientConfig & {
    maxKeyIterations?: number;
};
/**
 * NanoTDF Dataset SDK Client
 *
 *
 * @example
 * ```
 * import { clientSecretAuthProvider, NanoTDFDatasetClient } from '@opentdf/sdk';
 *
 * const OIDC_ENDPOINT = 'http://localhost:65432/auth/realms/opentdf';
 * const KAS_URL = 'http://localhost:65432/api/kas/';
 *
 * const ciphertext = '...';
 * const client = new NanoTDFDatasetClient({
 *   authProvider: await clientSecretAuthProvider({
 *     clientId: 'tdf-client',
 *     clientSecret: '123-456',
 *     exchange: 'client',
 *     oidcOrigin: OIDC_ENDPOINT,
 *   }),
 *   kasEndpoint: KAS_URL,
 * });
 * const plaintext = client.decrypt(ciphertext);
 * console.log('Plaintext', plaintext);
 * ```
 */
export declare class NanoTDFDatasetClient extends Client {
    static readonly NTDF_MAX_KEY_ITERATIONS = 8388606;
    private maxKeyIteration;
    private keyIterationCount;
    private cachedEphemeralKey?;
    private unwrappedKey?;
    private symmetricKey?;
    private cachedHeader?;
    private ecdsaBinding;
    /**
     * Create new NanoTDF Dataset Client
     *
     * The Ephemeral Key Pair can either be provided or will be generate when fetching the entity object. Once set it
     * cannot be changed. If a new ephemeral key is desired it a new client should be initialized.
     * There is no performance impact for creating a new client IFF the ephemeral key pair is provided.
     *
     * @param clientConfig OIDC client credentials
     * @param kasUrl Key access service URL
     * @param ephemeralKeyPair (optional) ephemeral key pair to use
     * @param maxKeyIterations Max iteration to performe without a key rotation
     */
    constructor(opts: DatasetConfig);
    /**
     * Encrypt data
     *
     * Pass a string, TypedArray, or ArrayBuffer data and get a promise which resolves ciphertext
     *
     * @param data to decrypt
     */
    encrypt(data: string | ArrayBufferLike, options?: EncryptOptions): Promise<ArrayBuffer>;
    /**
     * Decrypt ciphertext
     *
     * Pass a base64 string, TypedArray, or ArrayBuffer ciphertext and get a promise which resolves plaintext
     *
     * @param ciphertext Ciphertext to decrypt
     */
    decrypt(ciphertext: string | ArrayBufferLike): Promise<ArrayBuffer>;
    rewrapAndDecrypt(nanotdf: NanoTDF): Promise<ArrayBuffer>;
    generateIV(): Uint8Array;
}
//# sourceMappingURL=nanoclients.d.ts.map