/**
 * Computes an ECDSA signature for the given data using the provided private key.
 *
 * This function uses the Web Crypto API to generate a digital signature
 * for the input data using the ECDSA algorithm with SHA-256 as the hash function.
 *
 * @param {CryptoKey} privateKey - The ECDSA private key used for signing.
 * @param {Uint8Array} data - The data to be signed.
 * @returns {Promise<ArrayBuffer>} - A promise that resolves to the generated signature.
 */
export declare function computeECDSASig(privateKey: CryptoKey, data: Uint8Array): Promise<ArrayBuffer>;
/**
 * Verifies an ECDSA signature using the provided public key and data.
 *
 * This function uses the Web Crypto API to verify the digital signature
 * for the input data using the ECDSA algorithm with SHA-256 as the hash function.
 *
 * @param {CryptoKey} publicKey - The ECDSA public key used for verification.
 * @param {Uint8Array} signature - The signature to be verified.
 * @param {Uint8Array} data - The data that was signed.
 * @returns {Promise<boolean>} - A promise that resolves to a boolean indicating whether the signature is valid.
 */
export declare function verifyECDSASignature(publicKey: CryptoKey, signature: Uint8Array, data: Uint8Array): Promise<boolean>;
/**
 * Extracts the r and s values from a given ECDSA signature.
 *
 * @param {Uint8Array} signatureBytes - The raw ECDSA signature bytes.
 * @returns {{ r: Uint8Array; s: Uint8Array }} An object containing the r and s values as Uint8Arrays.
 * @throws {Error} If the validation of the signature fails.
 */
export declare function extractRSValuesFromSignature(signatureBytes: Uint8Array): {
    r: Uint8Array;
    s: Uint8Array;
};
//# sourceMappingURL=ecdsaSignature.d.ts.map