export type AssertionKeyAlg = 'ES256' | 'RS256' | 'HS256';
export type AssertionType = 'handling' | 'other';
export type Scope = 'tdo' | 'payload';
export type AppliesToState = 'encrypted' | 'unencrypted';
export type BindingMethod = 'jws';
export type Statement = {
    format: string;
    schema: string;
    value: string;
};
export type Binding = {
    method: string;
    signature: string;
};
export type Assertion = {
    id: string;
    type: AssertionType;
    scope: Scope;
    appliesToState?: AppliesToState;
    statement: Statement;
    binding: Binding;
};
export type AssertionPayload = {
    assertionHash: string;
    assertionSig: string;
};
/**
 * Computes the SHA-256 hash of the assertion object, excluding the 'binding' and 'hash' properties.
 *
 * @returns the hexadecimal string representation of the hash
 */
export declare function hash(a: Assertion): Promise<string>;
export declare function isAssertionConfig(obj: unknown): obj is AssertionConfig;
/**
 * Verifies the signature of the assertion using the provided key.
 *
 * @param {AssertionKey} key - The key used for verification.
 * @returns {Promise<[string, string]>} A promise that resolves to a tuple containing the assertion hash and signature.
 * @throws {Error} If the verification fails.
 */
export declare function verify(thiz: Assertion, aggregateHash: Uint8Array, key: AssertionKey, isLegacyTDF: boolean): Promise<void>;
/**
 * Creates an Assertion object with the specified properties.
 */
export declare function CreateAssertion(aggregateHash: Uint8Array, assertionConfig: AssertionConfig): Promise<Assertion>;
export type AssertionKey = {
    alg: AssertionKeyAlg;
    key: CryptoKey | Uint8Array;
};
export type AssertionConfig = {
    id: string;
    type: AssertionType;
    scope: Scope;
    appliesToState: AppliesToState;
    statement: Statement;
    signingKey?: AssertionKey;
};
export type AssertionVerificationKeys = {
    DefaultKey?: AssertionKey;
    Keys: Record<string, AssertionKey>;
};
//# sourceMappingURL=assertions.d.ts.map